#include <iostream>
#include <algorithm>
#include <map>
#include <set>
#include <sstream>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cassert>
#include <vector>
#include <string>
#include <fstream>

using namespace std;

const int c = 10;

struct Segment {
  string name, seq;
  int begin, end;

  Segment (){}
  Segment (const Segment &s) : name (s.name), seq (s.seq), begin (s.begin), end (s.end) {}
};

struct BP {
  int x, y, i, j;
  
  BP (){}
  BP (int x, int y, int i, int j) : x(x), y(y), i(i), j(j) {}  
  bool operator< (const BP &bp) const {
    if (x < bp.x) return true; else if (x > bp.x) return false;
    if (y < bp.y) return true; else if (y > bp.y) return false;
    if (i < bp.i) return true; else if (i > bp.i) return false;
    return (j < bp.j);
  }
};

vector<vector<Segment> > ref;
vector<vector<Segment> > test;

map<string,int> seqnum;

vector<vector<Segment> > read_ref_alignments (const char *filename){
  vector<vector<Segment> > res;
  vector<Segment> curr;

  ifstream data (filename);
  string s;

  while (getline (data, s)){
    if (s.length() == 0){
      if (curr.size() > 0) res.push_back (curr);
      curr.clear();
    } else {
      Segment d;
      istringstream iss (s);
      iss >> d.name >> d.begin >> d.seq >> d.end;
      curr.push_back (d);
    }    
  }
  if (curr.size() > 0) res.push_back (curr);
  data.close();

  return res;
}

vector<vector<Segment> > read_test_alignments (const char *filename){
  vector<vector<Segment> > res;
  vector<Segment> curr;

  ifstream data (filename);
  string s;

  while (data >> s){
    
    bool is_annotation = true;
    bool is_header = false;
    for (int i = 0; i < s.length(); i++){
      if (s[i] != ':' && s[i] != '*' && s[i] != '.') is_annotation = false;
      if (s[i] == '(') is_header = true;
    }
    
    if (is_annotation) continue;
    if (is_header){
      Segment d;
      int lparen = s.find ('(');
      int rparen = s.rfind (')');
      int hyphen = s.rfind ('-');
      d.name = s.substr(0, lparen);
      d.begin = atoi (s.substr(lparen+1, hyphen-lparen-1).c_str());
      d.end = atoi (s.substr(hyphen+1, rparen-hyphen-1).c_str());
      curr.push_back (d);
    } else {
      for (int i = 0; i < curr.size(); i++){
	if (i > 0) data >> s;
	data >> curr[i].seq;	
      }
      if (curr.size() > 0) res.push_back (curr);
      curr.clear();
    }
  }

  return res;
}

double agreement (double i, double j){
  double diff = fabs(i - j);
  if (diff <= c) return 1.0;
  if (diff >= 2*c) return 0.0;
  return 2.0 - diff / c;
}

double endpoint_agreement (const Segment &s, const Segment &t){
  if (s.name != t.name) return 0;
  return (agreement (s.begin, t.begin) + agreement (s.end, t.end)) / 2.0;
}

double midpoint_agreement (const Segment &s, const Segment &t){
  if (s.name != t.name) return 0;
  return agreement ((s.begin + s.end) / 2.0, (t.begin + t.end) / 2.0);
}

vector<Segment> set_union (const vector<vector<Segment> > &R){
  vector<Segment> ret;
  for (int i = 0; i < R.size(); i++)
    ret.insert (ret.end(), R[i].begin(), R[i].end());
  return ret;
}

double compute_segment_score (const vector<vector<Segment> > &R,
			      const vector<vector<Segment> > &T,
			      double (*agreement)(const Segment &, const Segment &)){
  
  const vector<Segment> A = set_union (R);
  const vector<Segment> B = set_union (T);
  
  double total = 0;
  for (int s = 0; s < A.size(); s++){
    double best_agreement = 0;
    for (int t = 0; t < B.size(); t++)
      best_agreement = max (best_agreement, agreement (A[s], B[t]));
    total += best_agreement;
  }
  return total / A.size();
}

bool overlap (const Segment &s, 
	      const Segment &t){

  if (s.name != t.name) return false;
  return (min (s.end, t.end) - max (s.begin, t.begin) + 1) >= 
    0.5 * max (s.end - s.begin + 1, t.end - t.begin + 1);
}

bool alignable_through (const Segment &s1,
			   const Segment &s2,
			   const vector<vector<Segment> > &T){

  for (int j = 0; j < T.size(); j++){
    for (int t1 = 0; t1 < T[j].size(); t1++) if (overlap (s1, T[j][t1])){
      for (int t2 = 0; t2 < T[j].size(); t2++) if (t1 != t2 && overlap (s2, T[j][t2])){
	return 1;
      }
    }
  }

  return 0;
}

double compute_cluster_score (const vector<vector<Segment> > &R,
			      const vector<vector<Segment> > &T){

  double score = 0;
  int ct = 0;
  
  for (int i = 0; i < R.size(); i++){
    for (int s1 = 0; s1 < R[i].size(); s1++){
      for (int s2 = s1+1; s2 < R[i].size(); s2++){
	score += alignable_through (R[i][s1], R[i][s2], T);
	ct++;
      }
    }
  }
  
  return score / ct;
}

set<BP> compute_matches (const Segment &s, const Segment &t){
  int s_seqnum = seqnum[s.name];
  int t_seqnum = seqnum[t.name];
  if (s_seqnum > t_seqnum) return compute_matches (t, s);
  
  set<BP> res;
  int s_pos = s.begin-1;
  int t_pos = t.begin-1;
  for (int i = 0; i < s.seq.length(); i++){
    if (isalpha(s.seq[i])) s_pos++;
    if (isalpha(t.seq[i])) t_pos++;
    if (isalpha(s.seq[i]) && isalpha(t.seq[i]))
      res.insert (BP(s_seqnum, t_seqnum, s_pos, t_pos));
  }
  return res;
}

set<BP> compute_matches (const vector<vector<Segment> > &R){
  set<BP> res;
  for (int i = 0; i < R.size(); i++){
    for (int j = 0; j < R[i].size(); j++){
      for (int k = j+1; k < R[i].size(); k++){
	set<BP> matches = compute_matches (R[i][j], R[i][k]);
	res.insert (matches.begin(), matches.end());
      }
    }
  }
  return res;
}

void add_names (const vector<vector<Segment> > &R){
  static set<string> used_names;
  for (int i = 0; i < R.size(); i++){
    for (int j = 0; j < R[i].size(); j++){
      if (used_names.find (R[i][j].name) == used_names.end()){
	used_names.insert (R[i][j].name);
	seqnum[R[i][j].name] = used_names.size();
      }
    }
  }
}

set<BP> intersect (set<BP> &X, set<BP> &Y){
  set<BP> ret;
  set_intersection (X.begin(), X.end(), Y.begin(), Y.end(),
		    inserter (ret, ret.begin()));
  return ret;
}

int main (int argc, char **argv){
  if (argc != 3){
    cerr << "Usage: score_alignments ref_file test_file" << endl;
    exit(1);
  }
  
  ref = read_ref_alignments (argv[1]);
  test = read_test_alignments (argv[2]);

  cout << "ref=" << argv[1] << "; "
       << "test=" << argv[2] << "; "
       << "epsn=" << compute_segment_score (ref, test, endpoint_agreement) << "; "
       << "epsp=" << compute_segment_score (test, ref, endpoint_agreement) << "; "
       << "mpsn=" << compute_segment_score (ref, test, midpoint_agreement) << "; "
       << "mpsp=" << compute_segment_score (test, ref, midpoint_agreement) << "; "
       << "csn=" << compute_cluster_score (ref, test) << "; "
       << "csp=" << compute_cluster_score (test, ref) << "; ";
  
  add_names (ref);
  add_names (test);

  set<BP> R_matches = compute_matches (ref);
  set<BP> T_matches = compute_matches (test);
  set<BP> RT_matches = intersect (R_matches, T_matches);
  
  cout << "rsn=" << (double) RT_matches.size() / R_matches.size() << "; "
       << "rsp=" << (double) RT_matches.size() / T_matches.size() << "; " << endl;

}
